from typing import Literal


DxMultiCCSName = Literal["dx_ccs"]  # diagnostics
PrMultiCCSName = Literal["pr_ccs"]  # procedures
DxFlatCCSName = Literal["dx_flat_ccs"]  # diagnostic
PrFlatCCSName = Literal["pr_flat_ccs"]  # procedures

ICD9CMName = Literal["icd9cm"]  # diagnostic
ICD9PCSName = Literal["icd9pcs"]  # procedures
ICD10CMName = Literal["icd10cm"]  # diagnostics
ICD10PCSName = Literal["icd10pcs"]  # procedures
