ADMISSION_TABLES_MAX_STAY_DAYS = 356
ADMISSION_CONCEPT_MAX_STAY_DAYS = 10
OBS_MAX_N_TIMESTAMPS = 1000
INTERVENTIONS_MAX_N_ITEMS = 50
ADMISSION_CONCEPT_MAX_STAY_HOURS = ADMISSION_CONCEPT_MAX_STAY_DAYS * 24
